import time
import argparse
import os
import subprocess
import shlex
import ruamel
from ruamel.yaml import YAML
from io import StringIO


def dump_to_string(obj, **kwargs):
    yaml = YAML(typ='unsafe', pure=True)
    with StringIO() as stream:
        yaml.dump(obj, stream, **kwargs)
        s = stream.getvalue()
    return s


def runcmd(cmd, *cmd_args, **subprocess_args):
    cmd = shlex.split(cmd) + list(cmd_args)
    #print(" ".join([f"'{a}'" for a in cmd]), flush=True)
    proc = subprocess.run(cmd, **subprocess_args)
    return proc


def getoutput(cmd, *cmd_args, **subprocess_args):
    proc = runcmd(cmd, *cmd_args, **subprocess_args, check=True,
                  stdout=subprocess.PIPE)
    return proc.stdout.decode("utf8")


def start_build(args):
    ts = time.time()
    with open(args.out, "w") as f:
        f.write(str(ts))


def finish_build(args):
    ts = time.time()
    with open(args.out, "r") as f:
        start = float(f.read())
    duration = ts - start
    results = {
        'compile': f"{duration:.3f}s",
        'file_size': f"{os.path.getsize(args.exe)}B"
    }
    s = dump_to_string({args.target: results})
    print(s, flush=True, end="")
    ## too much output:
    #if args.unix:
    #    # https://stackoverflow.com/questions/35485
    #    results['size'] = getoutput('size', args.exe)
    #    #results['symbols'] = getoutput('nm -t d -l -S --size-sort', args.exe)
    #s = dump_to_string({args.target: results})
    with open(args.out, "w") as f:
        f.write(s)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    #
    sp = subparsers.add_parser("start")
    sp.set_defaults(func=start_build)
    sp.add_argument('target', type=str, help='the target name')
    #
    sp = subparsers.add_parser("finish")
    sp.set_defaults(func=finish_build)
    sp.add_argument('target', type=str, help='the target name')
    sp.add_argument('exe', type=str, help='the executable file')
    sp.add_argument('-u', '--unix', action="store_true", help='use unix style size reporters')
    #
    args = parser.parse_args()
    args.out = f"{args.target}.dat"
    args.func(args)





def nb(val):
    return [val-1, val, val+1]


def ipowers2(min, max):
    vals = []
    v = int(min / 2)
    while v < -10:
        vals += nb(v)
        v = int(v / 2)
    vals += upowers2(max)
    return vals


def upowers2(max):
    vals = []
    v = 16
    while v < max:
        vals += nb(v)
        v *= 2
    return vals


def ipowers10(min, max):
    vals = []
    v = -100
    while v > min:
        vals += nb(v)
        v *= 10
    vals += upowers10(max)
    return vals


def upowers10(max):
    vals = []
    v = 10
    while v < max:
        vals += nb(v)
        v *= 10
    return vals


def idiv10(min, max):
    vals = []
    v = int(min / 10)
    while v < -10:
        vals.append(v)
        v = int(v / 10)
    vals += udiv10(max)
    return vals


def udiv10(max):
    vals = []
    v = int(max / 10)
    while v > 10:
        vals += nb(v)
        v = int(v / 10)
    return vals


def ivals(bits):
    min = -(1 << (bits-1))
    max = -min-1
    vals = [min, min+1, min+2, min+3, min+4, min+5]
    vals += ipowers2(min, max)
    vals += ipowers10(min, max)
    vals += idiv10(min, max)
    vals += [-11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    vals += [max-5, max-4, max-3, max-2, max-1, max]
    return sorted(list(set(vals)))


def uvals(bits):
    max = (1 << bits) - 1
    vals = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    vals += upowers2(max)
    vals += upowers10(max)
    vals += udiv10(max)
    vals += [max-5, max-4, max-3, max-2, max-1, max]
    return sorted(list(set(vals)))



def p(v):
    return f'        nc({v}, "{v}", "{hex(v)}", "{oct(v)}", "{bin(v)}"),'


def pn(numbits, fn):
    print()
    for a in fn(numbits):
        print(p(a))


pn(8, ivals)
pn(8, uvals)

pn(16, ivals)
pn(16, uvals)

pn(32, ivals)
pn(32, uvals)

pn(64, ivals)
pn(64, uvals)



import sys
import os
import re

thisdir = os.path.dirname(os.path.abspath(__file__))
moddir = os.path.abspath(f"{thisdir}/../cmake/bm-xp")
sys.path.insert(0, moddir)
import bm_plot as bm
from bm_util import first

from dataclasses import dataclass
import prettytable


def get_function_benchmark(function_name, run: bm.BenchmarkRun):
    for rbm in run.entries:
        if rbm.meta.function == function_name:
            return rbm
    raise Exception(f"function not found: {function_name}. Existing: {[rbm.meta.function for rbm in run.entries]}")


@dataclass
class CharconvMeta:  # also for atox
    title: str
    subject: str
    function: str
    data_type: bm.FundamentalTypes

    @classmethod
    def make(cls, bm_title: str):
        # eg:
        #   xtoa_c4_write_dec<uint8_t>
        #   xtoa_c4_utoa<uint8_t>
        #   xtoa_c4_xtoa<uint8_t>
        #   xtoa_c4_to_chars<uint8_t>
        #   xtoa_std_to_chars<uint8_t>
        #   xtoa_std_to_string<uint8_t>
        #   xtoa_sprintf<uint8_t>
        #   xtoa_sstream_reuse<uint8_t>
        #   xtoa_sstream<uint8_t>
        rx = re.compile(r'(atox|xtoa|xtoahex|xtoaoct|xtoabin)_(.*?)<(u?int\d+_t|float|double)>')
        if not rx.fullmatch(bm_title):
            raise Exception(f"cannot understand bm title: {bm_title}")
        subject = rx.sub(r'\1', bm_title)
        function = rx.sub(r'\2', bm_title)
        data_type = rx.sub(r'\3', bm_title)
        return cls(
            title=bm_title,
            subject=subject,
            function=function.replace("c4_", "c4::").replace("std_", "std::"),
            data_type=bm.FundamentalTypes.make(data_type)
        )

    @property
    def shortname(self):
        return self.function

    @property
    def shortparams(self):
        return str(self.data_type.short)

    @property
    def shorttitle(self):
        return f"{self.shortname}<{self.shortparams}>"


@dataclass
class CharconvThreadsMeta:
    function: str
    num_threads: int

    @classmethod
    def make(cls, bm_title: str):
        # eg:
        #   c4_itoa/real_time/threads:4
        rx = re.compile(r'(.*?)/real_time/threads:(\d+)')
        if not rx.fullmatch(bm_title):
            raise Exception(f"cannot understand bm title: {bm_title}")
        function = rx.sub(r'\1', bm_title)
        num_threads = int(rx.sub(r'\2', bm_title))
        return cls(
            function=function.replace("c4_", "c4::").replace("std_", "std::"),
            num_threads=num_threads
        )

    def checkbox_groups(self):
        return {}

    @property
    def shortname(self):
        return self.function

    @property
    def shorttitle(self):
        return self.shortname


def plot_charconv_bars(bm_panel: bm.BenchmarkPanel, getref,
                       panel_title_human: str,
                       outputfile_prefix: str):
    assert os.path.isabs(outputfile_prefix), outputfile_prefix
    # make a comparison table
    anchor = lambda run: f"{os.path.basename(outputfile_prefix)}-{first(run.meta).data_type}"
    anchorlink = lambda run: f"<pre><a href='#{anchor(run)}'>{first(run.meta).data_type}</a></pre>"
    with open(f"{outputfile_prefix}.txt", "w") as tablefile:
        with open(f"{outputfile_prefix}.md", "w") as mdfile:
            print(f"## {panel_title_human}\n\n<p>Data type benchmark results:</p>\n<ul>\n",
                  "\n".join([f"  <li>{anchorlink(run)}</li>" for run in bm_panel.runs]),
                  "</ul>\n\n", file=mdfile)
            for run in bm_panel.runs:
                data_type = first(run.meta).data_type
                tabletitle = f"{outputfile_prefix}-{data_type.short}"
                table = prettytable.PrettyTable(title=f"{panel_title_human}: {data_type}")
                table.add_column("function", [m.shorttitle for m in run.meta], align="l")
                for prop in ("mega_bytes_per_second", "cpu_time_ms"):
                    ref = getref(run)
                    bar_values = list(run.extract_plot_series(prop))
                    bar_values_rel = list(run.extract_plot_series(prop, relative_to_entry=ref))
                    bar_values_pc = list(run.extract_plot_series(prop, percent_of_entry=ref))
                    pd = bm_panel.first_run.property_plot_data(prop)
                    hns = pd.human_name_short
                    table.add_column(hns, [f"{v_:7.2f}" for v_ in bar_values], align="r")
                    hns = hns.replace(" (ms)", "")
                    table.add_column(f"{hns}(x)", [f"{v_:5.2f}x" for v_ in bar_values_rel], align="r")
                    table.add_column(f"{hns}(%)", [f"{v_:7.2f}%" for v_ in bar_values_pc], align="r")
                print(table, "\n\n")
                print(table, "\n\n", file=tablefile)
                pfx_bps = f"{os.path.basename(outputfile_prefix)}-mega_bytes_per_second-{data_type.short}"
                pfx_cpu = f"{os.path.basename(outputfile_prefix)}-cpu_time_ms-{data_type.short}"
                print(f"""
<br/>
<br/>

---

<a id="{anchor(run)}"/>

### {panel_title_human}: `{data_type}`

* Interactive html graphs for `{data_type}`:
  * [MB/s](./{pfx_bps}.html)
  * [CPU time](./{pfx_cpu}.html)

[![{data_type}: MB/s](./{pfx_bps}.png)](./{pfx_bps}.png)
[![{data_type}: CPU time](./{pfx_cpu}.png)](./{pfx_cpu}.png)

```
{table}
```
""", file=mdfile)
    # make plots
    for prop in ("mega_bytes_per_second", "cpu_time_ms"):
        ps, ps_ = [], []
        pd = bm_panel.first_run.property_plot_data(prop)
        bar_label = f"{pd.human_name_short}{pd.qty_type.comment}"
        outfilename = f"{outputfile_prefix}-{prop}"
        for run in bm_panel.runs:
            data_type = first(run.meta).data_type
            bar_names = [m.shorttitle for m in run.meta]
            bar_values = list(run.extract_plot_series(prop))
            runtitle = f"{outfilename}-{data_type.short}"
            # to save each bokeh plot separately and also
            # a grid plot with all of them, we have to plot
            # twice because bokeh does not allow saving twice
            # the same plot from multiple pictures.
            plotit = lambda: bm.plot_benchmark_run_as_bars(run, title=f"{panel_title_human}: {data_type}\n{bar_label}",
                                                           bar_names=bar_names, bar_values=bar_values, bar_label=bar_label)
            # make one plot to save:
            p, p_ = plotit()
            bm._bokeh_save_html(f"{runtitle}.html", p)
            bm._plt_save_png(f"{runtitle}.png")
            bm._plt_clear()
            # and another to gather:
            p, p_ = plotit()
            ps.append(p)
            ps_.append(p_)
            bm._plt_clear()
        bm.bokeh_plot_many(ps, f"{outfilename}.html")


def plot_itoa_threads_(bm_panel: bm.BenchmarkPanel, getref,
                       panel_title_human: str,
                       outputfile_prefix: str):
    assert os.path.isabs(outputfile_prefix), outputfile_prefix
    orig = lambda yprop, **kw: lambda run: list(run.extract_plot_series(yprop, **kw))
    divnt = lambda yprop, **kw: lambda run: [v / n for v, n in run.extract_plot_series_with_threads(yprop, **kw)]
    mulnt = lambda yprop, **kw: lambda run: [v * n for v, n in run.extract_plot_series_with_threads(yprop, **kw)]
    xprop = "threads"
    xpd = bm_panel.first_run.property_plot_data(xprop)
    xlabel = f"{xpd.human_name_short}"
    for yprop, ylog, yget in (
            #("mega_items_per_second", False, orig),
            ("mega_bytes_per_second", False, orig),
            #("iterations", False, divnt),
            #("real_time_ms", True, mulnt),
            ("cpu_time_ms", True, orig),):
        ypd = bm_panel.first_run.property_plot_data(yprop)
        ylabel = f"{ypd.human_name_short}{ypd.qty_type.comment}"
        p = bm.plot_benchmark_panel_as_lines(
            bm_panel, f"{panel_title_human}\n{ylabel}",
            xget=orig("threads"),
            yget=yget(yprop),
            nameget=lambda run: first(run.meta).function,
            ylog=ylog,
            xlabel=xlabel,
            ylabel=ylabel
        )
        name = f"{outputfile_prefix}-lines-{yprop}"
        # save png using matplotlib
        bm._plt_save_png(f"{name}.png")
        bm._plt_clear()
        # save html using bokeh
        bm._bokeh_save_html(f"{name}.html", p)
        #bkp.show(p)
    return p


def plot_itoa_threads(dir_: str, json_files):
    panel = bm.BenchmarkPanel(json_files, CharconvThreadsMeta)
    ref = lambda bmrun: get_function_benchmark("std::to_chars", run=bmrun)
    plot_itoa_threads_(panel, ref,
                       f"itoa benchmark: convert 2M 32b integers to string",
                       f"{dir_}/c4core-bm-charconv_threads")


def plot_charconv_xtoa(dir_: str, json_files, is_ftoa: bool):
    fcase = "ftoa" if is_ftoa else "xtoa"
    panel = bm.BenchmarkPanel(json_files, CharconvMeta)
    ref = lambda bmrun: get_function_benchmark("sprintf", run=bmrun)
    plot_charconv_bars(panel, ref,
                       f"xtoa benchmark: convert 1M numbers to strings",
                       f"{dir_}/c4core-bm-charconv-{fcase}")


def plot_charconv_atox(dir_: str, json_files, is_atof: bool):
    fcase = "atof" if is_atof else "atox"
    panel = bm.BenchmarkPanel(json_files, CharconvMeta)
    ref = lambda bmrun: get_function_benchmark("scanf", run=bmrun)
    plot_charconv_bars(panel, ref,
                       f"atox benchmark: convert 1M strings to numbers",
                       f"{dir_}/c4core-bm-charconv-{fcase}")


if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) < 2:
        raise Exception(f"usage: {sys.executable} {sys.argv[0]} <atox|xtoa|itoa_threads|format|digits> benchmarkfile.json[,benchmarkfile2.json,...]")
    cmd = args[0]
    json_files = args[1:]
    dir_ = os.path.dirname(json_files[0])
    for jf in json_files:
        print("jf:", jf, flush=True)
        assert os.path.dirname(jf) == dir_, (os.path.dirname(jf), dir_)
        assert os.path.exists(jf), jf
    if cmd == "itoa_threads":
        plot_itoa_threads(dir_, json_files)
    elif cmd == "xtoa" or cmd == "ftoa":
        plot_charconv_xtoa(dir_, json_files, (cmd == "ftoa"))
    elif cmd == "atox" or cmd == "atof":
        plot_charconv_atox(dir_, json_files, (cmd == "atof"))
    elif cmd == "format":
        raise Exception(f"not implemented: {cmd}")
    elif cmd == "digits":
        pass  # nothing to do
    else:
        raise Exception(f"not implemented: {cmd}")



import re
from os.path import abspath, dirname
import sys
import subprocess

projdir = abspath(dirname(dirname(__file__)))
sys.path.insert(0, f"{projdir}/cmake")
import amalgamate_utils as am


def amalgamate_fastfloat():
    fastfloatdir = f"{projdir}/src/c4/ext/fast_float"
    subprocess.run([
        sys.executable,
        f"{fastfloatdir}/script/amalgamate.py",
        "--license", "MIT",
        "--output", f"{fastfloatdir}/../fast_float_all.h"
    ], cwd=fastfloatdir).check_returncode()


def amalgamate_c4core(filename: str,
                      with_stl: bool=True,
                      with_fastfloat: bool=True):
    if with_fastfloat:
        amalgamate_fastfloat()
    repo = "https://github.com/biojppm/c4core"
    defmacro = "C4CORE_SINGLE_HDR_DEFINE_NOW"
    exports_def_code = f"""// shared library: export when defining
#if defined(C4CORE_SHARED) && defined({defmacro}) && !defined(C4CORE_EXPORTS)
#define C4CORE_EXPORTS
#endif
"""
    required_gcc4_8_include = """// (amalgamate) these includes are needed to work around
// conditional includes in the gcc4.8 shim
#include <cstdint>
#include <type_traits>
#include <cstring>
"""
    required_charconv_include = """// (amalgamate) this include is needed to work around
// conditional includes in charconv.hpp
#if (defined(_MSVC_LANG) && (_MSVC_LANG >= 201703L)) || (__cplusplus >= 201703L)
#include <charconv>
#endif
"""
    srcblocks = [
        am.cmttext(f"""
c4core - C++ utilities

{repo}

DO NOT EDIT. This file is generated automatically.
This is an amalgamated single-header version of the library.

INSTRUCTIONS:
  - Include at will in any header of your project
  - In one (and only one) of your project source files,
    #define {defmacro} and then include this header.
    This will enable the function and class definitions in
    the header file.
  - To compile into a shared library, just define the
    preprocessor symbol C4CORE_SHARED . This will take
    care of symbol export/import.
"""),
        am.cmtfile("LICENSE.txt"),
        am.injcode(exports_def_code),
        "src/c4/export.hpp",
        "src/c4/version.hpp",
        "src/c4/preprocessor.hpp",
        "src/c4/platform.hpp",
        "src/c4/cpu.hpp",
        am.injcode(required_gcc4_8_include),
        "src/c4/gcc-4.8.hpp",
        "src/c4/compiler.hpp",
        "src/c4/language.hpp",
        "src/c4/types.hpp",
        "src/c4/config.hpp",
        am.hdrfile("src/c4/ext/debugbreak/debugbreak.h", "c4/ext/debugbreak/debugbreak.h", "DEBUG_BREAK_H"),
        "src/c4/error.hpp",
        "src/c4/memory_util.hpp",
        "src/c4/memory_resource.hpp",
        "src/c4/ctor_dtor.hpp",
        "src/c4/allocator.hpp",
        "src/c4/char_traits.hpp",
        "src/c4/hash.hpp",
        "src/c4/szconv.hpp",
        "src/c4/blob.hpp",
        "src/c4/substr_fwd.hpp",
        "src/c4/substr.hpp",
        am.onlyif(with_fastfloat, am.injfile("src/c4/ext/fast_float_all.h", "c4/ext/fast_float_all.h")),
        am.onlyif(with_fastfloat, "src/c4/ext/fast_float.hpp"),
        "src/c4/std/vector_fwd.hpp",
        "src/c4/std/string_fwd.hpp",
        "src/c4/std/std_fwd.hpp",
        am.injcode(required_charconv_include),
        "src/c4/charconv.hpp",
        "src/c4/utf.hpp",
        "src/c4/format.hpp",
        "src/c4/dump.hpp",
        "src/c4/enum.hpp",
        "src/c4/bitmask.hpp",
        "src/c4/span.hpp",
        "src/c4/type_name.hpp",
        "src/c4/base64.hpp",
        am.onlyif(with_stl, am.ignfile("src/c4/std/std.hpp")), # this is an umbrella include
        am.onlyif(with_stl, "src/c4/std/string.hpp"),
        am.onlyif(with_stl, "src/c4/std/string_view.hpp"),
        am.onlyif(with_stl, "src/c4/std/vector.hpp"),
        am.onlyif(with_stl, "src/c4/std/tuple.hpp"),
        "src/c4/ext/rng/rng.hpp",
        "src/c4/ext/sg14/inplace_function.h",
        am.ignfile("src/c4/common.hpp"),
        am.ignfile("src/c4/c4_push.hpp"),
        am.ignfile("src/c4/c4_pop.hpp"),
        am.ignfile("src/c4/restrict.hpp"),
        am.ignfile("src/c4/unrestrict.hpp"),
        "src/c4/version.cpp",
        "src/c4/language.cpp",
        "src/c4/format.cpp",
        "src/c4/memory_util.cpp",
        "src/c4/char_traits.cpp",
        "src/c4/memory_resource.cpp",
        "src/c4/utf.cpp",
        "src/c4/base64.cpp",
        am.injcode("#define C4_WINDOWS_POP_HPP_"),
        "src/c4/windows_push.hpp",
        "src/c4/windows.hpp",
        "src/c4/windows_pop.hpp", # do NOT include this before windows.hpp
        "src/c4/error.cpp",
    ]
    result = am.catfiles(srcblocks,
                         projdir,
                         # comment out lines with these patterns:
                         include_regexes=[
                             re.compile(r'^\s*#\s*include "(c4/.*)".*$'),
                             re.compile(r'^\s*#\s*include <(c4/.*)>.*$'),
                         ],
                         definition_macro=defmacro,
                         repo=repo,
                         result_incguard="_C4CORE_SINGLE_HEADER_AMALGAMATED_HPP_")
    result_with_only_first_includes = am.include_only_first(result)
    am.file_put_contents(filename, result_with_only_first_includes)


def mkparser():
    return am.mkparser(fastfloat=(True, "enable fastfloat library"),
                       stl=(True, "enable stl interop"))


if __name__ == "__main__":
    args = mkparser().parse_args()
    amalgamate_c4core(filename=args.output,
                      with_fastfloat=args.fastfloat,
                      with_stl=args.stl)
